r"""Base modules

In Inox, a module is a PyTree whose branches are its attributes. A branch can be any
PyTree-compatible object (:py:`bool`, :py:`str`, :py:`list`, :py:`dict`, ...), including
other modules. Parametric functions, such as neural networks, should subclass
:class:`Module` and indicate their parameters with :class:`Parameter`.

.. code-block:: python

    import inox
    import inox.nn as nn
    import jax
    import jax.numpy as jnp

    class Linear(nn.Module):
        def __init__(self, in_features, out_features, key):
            keys = jax.random.split(key, 2)

            self.weight = nn.Parameter(jax.random.normal(keys[0], (in_features, out_features)))
            self.bias = nn.Parameter(jax.random.normal(keys[1], (out_features,)))

        def __call__(self, x):
            return x @ self.weight() + self.bias()

    class Classifier(nn.Module):
        def __init__(self, in_features, num_classes, key):
            keys = jax.random.split(key, 3)

            self.l1 = Linear(in_features, 64, key=keys[0])
            self.l2 = Linear(64, 64, key=keys[1])
            self.l3 = Linear(64, num_classes, key=keys[2])
            self.relu = nn.ReLU()

            self.return_logits = True

        def __call__(self, x):
            x = self.l1(x)
            x = self.l2(self.relu(x))
            x = self.l3(self.relu(x))

            if self.return_logits:
                return x
            else:
                return jax.nn.softmax(x)

    key = jax.random.key(0)
    model = Classifier(16, 3, key)
"""

from __future__ import annotations

__all__ = [
    "Module",
    "ModuleDef",
    "Parameter",
    "ComplexParameter",
]

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from jax import Array
from typing import Any, Callable, Dict, NamedTuple, Tuple, Union

import inox.tree


def is_module(x: Any) -> bool:
    return isinstance(x, Module)


class Module(inox.tree.Namespace):
    r"""Base class for all modules.

    Arguments:
        kwargs: A name-value mapping.
    """

    def train(self, mode: bool = True):
        r"""Toggles between training and evaluation modes.

        This method is primarily useful for (sub)modules that behave differently at
        training and evaluation, such as :class:`inox.nn.dropout.TrainingDropout` and
        :class:`inox.nn.normalization.BatchNorm`.

        Arguments:
            mode: Whether to turn training mode on or off.

        Example:
            >>> model = Module(dropout=nn.TrainingDropout())
            >>> model
            Module(
              dropout = TrainingDropout(
                p = float32[]
              )
            )
            >>> model.train(False)
            >>> model
            Module(
              dropout = TrainingDropout(
                p = float32[],
                training = False
              )
            )
        """

        for leaf in jtu.tree_leaves(vars(self), is_module):
            if is_module(leaf):
                leaf.train(mode)

        if hasattr(self, "training"):
            self.training = mode

    def partition(
        self,
        *filters: Union[type, Callable[[Any], bool]],
    ) -> Tuple[ModuleDef, Dict[str, Array]]:
        r"""Splits the static definition of the module from its arrays.

        The arrays are partitioned into a set of path-array mappings. The mapping in
        which an array is contained is chosen according to its oldest (closest to the
        root) ancestor that satisfies a constraint. If a node satisfies several
        constraints, the first one is selected. The last mapping is dedicated to arrays
        that do not satisfy any constraint.

        See also:
            :class:`inox.tree.partition`

        Arguments:
            filters: A set of filtering constraints. Types are transformed into
                :py:`isinstance` constraints.

        Returns:
            The module static definition and array partitions.

        Examples:
            >>> keys = jax.random.split(jax.random.key(0), 2)
            >>> model = Module(layers=[
            ...     nn.Linear(3, 64, key=keys[0]),
            ...     nn.ReLU(),
            ...     nn.TrainingDropout(),
            ...     nn.Linear(64, 5, key=keys[1]),
            ... ])
            >>> static, arrays = model.partition()
            >>> print(inox.tree.prepr(arrays))
            {
              '.layers[0].bias.value': float32[64],
              '.layers[0].weight.value': float32[3, 64],
              '.layers[2].p': float32[],
              '.layers[3].bias.value': float32[5],
              '.layers[3].weight.value': float32[64, 5]
            }

            >>> static, params, others = model.partition(nn.Parameter)
            >>> print(inox.tree.prepr(params))
            {
              '.layers[0].bias.value': float32[64],
              '.layers[0].weight.value': float32[3, 64],
              '.layers[3].bias.value': float32[5],
              '.layers[3].weight.value': float32[64, 5]
            }
            >>> print(inox.tree.prepr(others))
            {'.layers[2].p': float32[]}

            >>> grads = jax.tree.map(jax.numpy.ones_like, params)
            >>> params = jax.tree.map(lambda x, y: x + 0.01 * y, params, grads)
            >>> model = static(params, others)  # updated copy

            >>> model.layers[3].frozen = True
            >>> filtr = lambda x: getattr(x, 'frozen', False)
            >>> static, frozen, params, others = model.partition(filtr, nn.Parameter)
            >>> print(inox.tree.prepr(frozen))
            {'.layers[3].bias.value': float32[5], '.layers[3].weight.value': float32[64, 5]}
            >>> print(inox.tree.prepr(params))
            {'.layers[0].bias.value': float32[64], '.layers[0].weight.value': float32[3, 64]}
        """

        treedef, *leaves = inox.tree.partition(self, *filters)

        static = {}
        arrays = [{} for _ in leaves]

        for i, partition in enumerate(leaves):
            for path, leaf in partition.items():
                if inox.tree.is_array(leaf):
                    arrays[i][path] = leaf
                else:
                    static[path] = leaf

        return ModuleDef(treedef, static), *arrays


class ModuleDef(NamedTuple):
    r"""Abstraction for the static definition of a module.

    See also:
        :meth:`Module.partition`

    Arguments:
        treedef: A module tree definition.
        leaves: The static (non-array) leaves of the module.
    """

    treedef: inox.tree.PyTreeDef
    leaves: Dict[str, Any]

    def __call__(self, *arrays: Dict[str, Array]) -> Module:
        r"""
        Arguments:
            arrays: A set of array partitions.

        Returns:
            A new instance of the module.
        """

        return inox.tree.combine(self.treedef, self.leaves, *arrays)


class Parameter(inox.tree.Namespace):
    r"""Wrapper to indicate an optimizable array.

    All arrays that require gradient updates in a :class:`Module` should be wrapped in a
    :class:`Parameter` instance.

    Arguments:
        value: An array.

    Example:
        >>> weight = Parameter(jax.numpy.ones((3, 5))); weight
        Parameter(float32[3, 5])
        >>> x = jax.random.normal(jax.random.key(0), (16, 3))
        >>> y = x @ weight()
    """

    value: Array = None

    def __init__(self, value: Array):
        self.value = jnp.asarray(value)

    def __call__(self) -> Array:
        r"""
        Returns:
            The wrapped array.
        """

        return self.value

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.value, attr)

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return f"{self.__class__.__name__}({inox.tree.prepr(self.value, **kwargs)})"


class ComplexParameter(Parameter):
    r"""Wrapper to indicate an optimizable complex array.

    The real and imaginary parts are stored as separate floating point arrays to enable
    gradient-based optimization.

    Arguments:
        value: A complex array.

    Example:
        >>> value = jax.numpy.ones((3, 5)) + 1j * jax.numpy.zeros((3, 5))
        >>> weight = ComplexParameter(value); weight
        ComplexParameter(complex64[3, 5])
        >>> x = jax.random.normal(jax.random.key(0), (16, 3))
        >>> y = x @ weight()
    """

    real: Array = None
    imag: Array = None

    def __init__(self, value: Array):
        value = jnp.asarray(value)

        self.real = value.real
        self.imag = value.imag

    @property
    def value(self) -> Array:
        try:
            return jax.lax.complex(self.real, self.imag)
        except TypeError:
            return None



r"""Extended utilities for random number generation."""

__all__ = [
    "PRNG",
    "set_rng",
    "get_rng",
]

import jax

from contextlib import contextmanager
from functools import partial
from jax import Array
from typing import Any, Dict, Union

from .debug import same_trace
from .tree import Namespace


class PRNG(Namespace):
    r"""Creates a pseudo-random number generator (PRNG).

    This class is a thin wrapper around the :mod:`jax.random` module, and allows to
    generate new PRNG keys or sample from distributions without having to split keys
    with :func:`jax.random.split` by hand.

    Arguments:
        seed: An integer seed or PRNG key.
        kwargs: Keyword arguments passed to :func:`jax.random.PRNGKey`.

    Example:
        >>> rng = PRNG(42)
        >>> rng.split()  # generates a key
        Array([1832780943,  270669613], dtype=uint32)
        >>> rng.split(3)  # generates a vector of 3 keys
        Array([[3187376881,  129218101],
               [2350016172, 1168365246],
               [ 257214496,  567757975]], dtype=uint32)
        >>> rng.normal((5,))
        Array([ 0.6611632 , -1.0414096 ,  0.5554834 , -1.8841821 ,  0.36664668],      dtype=float32)
    """

    def __init__(self, seed: Union[int, Array], **kwargs):
        if isinstance(seed, int):
            self.state = jax.random.PRNGKey(seed, **kwargs)
        else:
            self.state = seed

    def __getattr__(self, name: str) -> Any:
        attr = getattr(jax.random, name)

        if callable(attr):
            return partial(attr, self.split())
        else:
            return attr

    def split(self, num: int = None) -> Array:
        r"""
        Arguments:
            num: The number of keys to generate.

        Returns:
            A new key if :py:`num=None` and a vector of keys otherwise.
        """

        if num is None:
            keys = jax.random.split(self.state, num=2)
        else:
            keys = jax.random.split(self.state, num=num + 1)

        assert same_trace(self.state, keys), (
            "the RNG state was initialized and used within different JIT traces."
        )

        if num is None:
            key, self.state = keys
        else:
            key, self.state = keys[:-1], keys[-1]

        return key


INOX_RNG: Dict[str, PRNG] = {}


@contextmanager
def set_rng(**rngs: Dict[str, PRNG]):
    r"""Sets named RNG states within a context.

    See also:
        :func:`get_rng`

    Arguments:
        rngs: Named :class:`PRNG` instances.

    Example:
        >>> with set_rng(init=PRNG(0), dropout=PRNG(42)):
        ...     keys = get_rng("init").split(3)
        ...     mask = get_rng("dropout").bernoulli(shape=(2, 3))
    """

    old = {}
    nil = object()

    try:
        for name, rng in rngs.items():
            old[name], INOX_RNG[name] = INOX_RNG.get(name, nil), rng
        yield
    finally:
        for name, rng in old.items():
            if rng is nil:
                del INOX_RNG[name]
            else:
                INOX_RNG[name] = rng


def get_rng(name: str) -> PRNG:
    r"""Returns a context-bound RNG state given its name.

    See also:
        :func:`set_rng`

    Arguments:
        name: The RNG state's name.
    """

    assert name in INOX_RNG, f"no RNG state with name '{name}' is set in this context."

    return INOX_RNG[name]



r"""Normalization layers"""

__all__ = [
    "BatchNorm",
    "LayerNorm",
    "GroupNorm",
]

import jax
import jax.numpy as jnp

from einops import rearrange
from jax import Array
from typing import Dict, NamedTuple, Sequence, Tuple, Union

from .module import Module
from .state import StateEntry, update_state


class Statistics(NamedTuple):
    mean: Array
    var: Array


class BatchNorm(Module):
    r"""Creates a batch-normalization layer.

    .. math:: y = \frac{x - \mathbb{E}[x]}{\sqrt{\mathbb{V}[x] + \epsilon}}

    The mean and variance are calculated over the batch and spatial axes. During
    training, the layer keeps running estimates of the mean and variance, which are then
    used for normalization during evaluation. The update rule for a running average
    statistic :math:`\hat{s}` is

    .. math:: \hat{s} \gets \alpha \hat{s} + (1 - \alpha) s

    where :math:`s` is the statistic calculated for the current batch.

    References:
        | Accelerating Deep Network Training by Reducing Internal Covariate Shift (Ioffe et al., 2015)
        | https://arxiv.org/abs/1502.03167

    Arguments:
        channels: The number of channels :math:`C`.
        epsilon: A numerical stability term :math:`\epsilon`.
        momentum: The momentum :math:`\alpha \in [0, 1]` for the running estimates.
    """

    training: bool = True

    def __init__(
        self,
        channels: int,
        epsilon: Union[float, Array] = 1e-05,
        momentum: Union[float, Array] = 0.9,
    ):
        self.epsilon = jnp.asarray(epsilon)
        self.momentum = jnp.asarray(momentum)

        self.stats = StateEntry(
            Statistics(
                mean=jnp.zeros(channels),
                var=jnp.ones(channels),
            )
        )

    def __call__(self, x: Array, state: Dict) -> Tuple[Array, Dict]:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(N, *, C)`.
            state: The state dictionary.

        Returns:
            The output tensor :math:`y`, with shape :math:`(N, *, C)`, and the
            (updated) state dictionary.
        """

        if self.training:
            assert x.ndim > 1, "the input tensor is not batched."

            y = x.reshape(-1, x.shape[-1])
            mean = jnp.mean(y, axis=0)
            var = jnp.var(y, axis=0)

            stats = state[self.stats]
            stats = Statistics(
                mean=self.ema(stats.mean, jax.lax.stop_gradient(mean)),
                var=self.ema(stats.var, jax.lax.stop_gradient(var)),
            )

            state = update_state(state, {self.stats: stats})
        else:
            mean, var = state[self.stats]

        y = (x - mean) / jnp.sqrt(var + self.epsilon)

        return y, state

    def ema(self, x: Array, y: Array) -> Array:
        return self.momentum * x + (1 - self.momentum) * y


class LayerNorm(Module):
    r"""Creates a layer-normalization layer.

    .. math:: y = \frac{x - \mathbb{E}[x]}{\sqrt{\mathbb{V}[x] + \epsilon}}

    References:
        | Layer Normalization (Ba et al., 2016)
        | https://arxiv.org/abs/1607.06450

    Arguments:
        axis: The axis(es) over which the mean and variance are calculated.
        epsilon: A numerical stability term :math:`\epsilon`.
    """

    def __init__(
        self,
        axis: Union[int, Sequence[int]] = -1,
        epsilon: Union[float, Array] = 1e-05,
    ):
        self.axis = axis
        self.epsilon = jnp.asarray(epsilon)

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, C)`.
        """

        mean = jnp.mean(x, axis=self.axis, keepdims=True)
        var = jnp.var(x, axis=self.axis, keepdims=True)

        return (x - mean) / jnp.sqrt(var + self.epsilon)


class GroupNorm(LayerNorm):
    r"""Creates a group-normalization layer.

    References:
        | Group Normalization (Wu et al., 2018)
        | https://arxiv.org/abs/1803.08494

    Arguments:
        groups: The number of groups :math:`G` to separate channels into. If :math:`G = 1`,
            the layer is equivalent to :class:`LayerNorm`.
        epsilon: A numerical stability term :math:`\epsilon`.
    """

    def __init__(
        self,
        groups: int,
        epsilon: Union[float, Array] = 1e-05,
    ):
        super().__init__(axis=-1, epsilon=epsilon)

        self.groups = groups

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, C)`.
        """

        x = rearrange(x, "... (G D) -> ... G D", G=self.groups)
        x = super().__call__(x)
        x = rearrange(x, "... G D -> ... (G D)")

        return x



r"""Interpolation layers"""

__all__ = [
    "Resample",
]

import jax
import math

from functools import partial
from jax import Array
from typing import Sequence

from .module import Module
from ..numpy import vectorize


class Resample(Module):
    r"""Creates a spatial resampling layer.

    Arguments:
        factor: The resampling factor in each spatial axis.
        kwargs: Keyword arguments passed to :func:`jax.image.resize`.
    """

    def __init__(self, factor: Sequence[float], **kwargs):
        self.factor = factor
        self.kwargs = kwargs

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C)`,
            such that

            .. math:: H_i' = \left\lfloor r_i \times H_i \right\rfloor

            where :math:`r_i` is the resampling factor of the :math:`i`-th spatial axis.
        """

        assert x.ndim >= self.ndim

        shape = x.shape[-self.ndim :]
        shape = [math.floor(r * h) for r, h in zip(self.factor, shape)]

        resize = partial(jax.image.resize, shape=shape, **self.kwargs)
        resize = jax.vmap(resize, in_axes=-1, out_axes=-1)
        resize = vectorize(resize, ndims=self.ndim)

        return resize(x)

    @property
    def ndim(self) -> int:
        return len(self.factor) + 1



r"""Extended utilities for debugging."""

__all__ = [
    "same_trace",
]


from jax import Array
from jax.core import Tracer


def same_trace(x: Array, y: Array, ignore_primal: bool = False) -> bool:
    r"""Checks whether two arrays have the same trace source.

    Arguments:
        x: The first array.
        y: The second array.
        ignore_primal: Whether to ignore primal traces (:func:`jax.grad`).

    Example:
        >>> x, y = jax.numpy.zeros(2)
        >>> same_trace(x, y)
        True
        >>> jax.jit(lambda x, y: same_trace(x, y))(x, y)
        Array(True, dtype=bool)
        >>> jax.jit(lambda x: same_trace(x, y))(x)
        Array(False, dtype=bool)
    """

    if ignore_primal:
        while hasattr(x, "primal"):
            x = x.primal

        while hasattr(y, "primal"):
            y = y.primal

    if isinstance(x, Tracer) and isinstance(y, Tracer):
        return x._trace.main == y._trace.main
    elif isinstance(x, Tracer) or isinstance(y, Tracer):
        return False
    else:
        return True



r"""Extended NumPy interface."""

__all__ = [
    "flatten",
    "unflatten",
    "vectorize",
]

import jax
import jax.numpy as jnp

from functools import wraps
from jax import Array
from typing import Callable, Iterable, Sequence, Union


def flatten(x: Array, start: int = 0, stop: int = None) -> Array:
    r"""Flattens an axis range of an array.

    Arguments:
        x: An array.
        start: The start of the axis range to flatten.
        stop: The end of the axis range to flatten (excluded).
            If :py:`None`, :py:`x.ndim` is used instead.

    Returns:
        The flattened array.

    Example:
        >>> x = jax.numpy.zeros((2, 3, 5))
        >>> flatten(x, 0, 2).shape
        (6, 5)
    """

    if stop is None:
        stop = x.ndim

    return x.reshape(*x.shape[:start], -1, *x.shape[stop:])


def unflatten(x: Array, axis: int, shape: Sequence[int]) -> Array:
    r"""Reshapes an axis of an array.

    Arguments:
        x: An array.
        axis: The axis to reshape.
        shape: The shape of the reshaped axis.

    Returns:
        The array with the reshaped axis.

    Example:
        >>> x = jax.numpy.zeros((6, 5))
        >>> unflatten(x, 0, (2, 3)).shape
        (2, 3, 5)
    """

    return x.reshape(*x.shape[:axis], *shape, *x.shape[axis % x.ndim + 1 :])


def vectorize(
    f: Callable,
    ndims: Union[int, Sequence[int]],
    exclude: Iterable[int] = (),
) -> Callable:
    r"""Vectorizes a function with broadcasting.

    :func:`vectorize` is similar to :func:`jax.numpy.vectorize` except that it takes the
    number of core dimensions of arguments as signature instead of their shape.

    Arguments:
        f: A function to vectorize.
        ndims: The number of dimensions expected for each positional argument.
        exclude: A set of positive indices representing positional arguments that
            should not be vectorized.

    Returns:
        The vectorized function.

    Example:
        >>> mvp = vectorize(jax.numpy.dot, (2, 1))  # broadcasting matrix-vector product
        >>> A = jax.numpy.ones((5, 3))
        >>> x = jax.random.normal(jax.random.key(0), (16, 3))
        >>> y = mvp(A, x)
        >>> y.shape
        (16, 5)
    """

    exclude = set(exclude)

    @wraps(f)
    def wrapped(*args, **kwargs):
        others = [arg for i, arg in enumerate(args) if i in exclude]
        args = [jnp.asarray(arg) for i, arg in enumerate(args) if i not in exclude]

        def g(*args):
            ia, io = iter(args), iter(others)
            args = (next(io if i in exclude else ia) for i in range(len(args) + len(others)))

            return f(*args, **kwargs)

        if isinstance(ndims, int):
            shapes = [arg.shape[: arg.ndim - ndims] for arg in args]
        else:
            assert len(args) == len(ndims)
            assert all(0 <= ndim <= arg.ndim for arg, ndim in zip(args, ndims))

            shapes = [arg.shape[: arg.ndim - ndim] for arg, ndim in zip(args, ndims)]

        broadcast = jnp.broadcast_shapes(*shapes)
        squeezed = []

        for arg, shape in zip(args, shapes):
            axes = [i for i, size in enumerate(shape) if size == 1]
            squeezed.append(jnp.squeeze(arg, axes))

        for i, size in enumerate(reversed(broadcast), start=1):
            in_axes = [None if len(shape) < i or shape[-i] == 1 else 0 for shape in shapes]

            g = jax.vmap(g, in_axes=in_axes, axis_size=size)

        return g(*squeezed)

    return wrapped



r"""Tests for the inox.debug module."""

import jax
import jax.numpy as jnp

from functools import partial

from inox.debug import same_trace


def test_same_trace():
    x = jnp.zeros(1)
    y = jnp.ones(1)

    assert same_trace(x, y)
    assert jax.jit(same_trace)(x, y)
    assert not jax.jit(partial(same_trace, x))(y)



r"""Extended utilities for tree-like data structures."""

__all__ = [
    "Namespace",
    "Partial",
    "Static",
    "mask_static",
    "unmask_static",
    "partition",
    "combine",
    "prepr",
]

import dataclasses
import jax._src.tree_util as jtu
import numpy as np

from jax import Array
from textwrap import indent
from typing import Any, Callable, Dict, Hashable, Tuple, TypeVar, Union
from warnings import warn

PyTree = TypeVar("PyTree", bound=Any)
PyTreeDef = TypeVar("PyTreeDef")


def is_array(x: Any) -> bool:
    return isinstance(x, np.ndarray) or isinstance(x, Array)


class PyTreeMeta(type):
    r"""PyTree meta-class."""

    def __new__(cls, *args, **kwargs) -> type:
        cls = super().__new__(cls, *args, **kwargs)

        if hasattr(cls, "tree_flatten_with_keys"):
            jtu.register_pytree_with_keys_class(cls)
        else:
            jtu.register_pytree_node_class(cls)

        return cls


class Namespace(metaclass=PyTreeMeta):
    r"""PyTree class for name-value mappings.

    Arguments:
        kwargs: A name-value mapping.

    Example:
        >>> tree = Namespace(a=1, b='2'); tree
        Namespace(
          a = 1,
          b = '2'
        )
        >>> tree.c = [3, False]; tree
        Namespace(
          a = 1,
          b = '2',
          c = [3, False]
        )
        >>> jax.tree.leaves(tree)
        [1, '2', 3, False]
    """

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def __repr__(self) -> str:
        return prepr(self)

    def tree_repr(self, **kwargs) -> str:
        private = kwargs.get("private", False)

        lines = (
            f"{name} = {prepr(getattr(self, name), **kwargs)}"
            for name in sorted(self.__dict__.keys())
            if not name.startswith("_") or private
        )

        lines = ",\n".join(lines)

        if lines:
            lines = "\n" + indent(lines, "  ") + "\n"

        return f"{self.__class__.__name__}({lines})"

    def tree_flatten(self):
        if self.__dict__:
            names, values = zip(*sorted(self.__dict__.items()))
        else:
            names, values = (), ()

        return values, names

    def tree_flatten_with_keys(self):
        values, names = self.tree_flatten()
        keys = map(jtu.GetAttrKey, names)

        return list(zip(keys, values)), names

    @classmethod
    def tree_unflatten(cls, names, values):
        self = object.__new__(cls)
        self.__dict__ = dict(zip(names, values))

        return self


class Partial(Namespace):
    r"""A version of :class:`functools.partial` that is a PyTree.

    Arguments:
        func: A function.
        args: Positional arguments for future calls.
        kwds: Keyword arguments for future calls.

    Examples:
        >>> increment = Partial(jax.numpy.add, 1)
        >>> increment(2)
        Array(3, dtype=int32, weak_type=True)

        >>> println = Partial(print, sep='\n')
        >>> println('Hello', 'World!')
        Hello
        World!
    """

    def __init__(self, func: Callable, *args: Any, **kwds: Any):
        self.func = func
        self.args = args
        self.kwds = kwds

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.func(*self.args, *args, **self.kwds, **kwds)


class Static(metaclass=PyTreeMeta):
    r"""Wraps an hashable value as a leafless PyTree.

    Arguments:
        value: An hashable value to wrap.

    Example:
        >>> tree = Static((0, 'one', None))
        >>> tree.value
        (0, 'one', None)
        >>> jax.tree.leaves(tree)
        []
        >>> jax.tree.structure(tree)
        PyTreeDef(CustomNode(Static[(0, 'one', None)], []))
    """

    def __init__(self, value: Hashable):
        if not isinstance(value, Hashable):
            warn(f"considering a non-hashable object ('{type(value).__name__}') static could lead to frequent JIT recompilations.", stacklevel=2)  # fmt: off
        elif callable(value):
            if "<lambda>" in value.__qualname__ or "<locals>" in value.__qualname__:
                warn(f"considering a local function ('{value.__qualname__}') static could lead to frequent JIT recompilations.", stacklevel=2)  # fmt: off

        self.value = value

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.value == other.value

    def __hash__(self) -> int:
        return hash((type(self), self.value))

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return f"{self.__class__.__name__}({prepr(self.value, **kwargs)})"

    def tree_flatten(self):
        return (), self.value

    def tree_flatten_with_keys(self):
        return self.tree_flatten()

    @classmethod
    def tree_unflatten(cls, value, _):
        self = object.__new__(cls)
        self.value = value

        return self


def mask_static(
    tree: PyTree,
    is_static: Callable[[Any], bool] = None,
) -> PyTree:
    r"""Masks the static leaves of a tree.

    The structure of the tree remains unchanged, but leaves that are considered static
    are wrapped into a :class:`Static` instance, which hides them from
    :func:`jax.tree.leaves` and :func:`jax.tree.map`.

    See also:
        :func:`inox.tree.unmask_static`

    Arguments:
        tree: The tree to mask.
        is_static: A predicate for what to consider static. If :py:`None`,
            all non-array leaves are considered static.

    Returns:
        The masked tree.

    Example:
        >>> tree = [1, jax.numpy.arange(2), 'three']
        >>> jax.tree.leaves(tree)
        [1, Array([0, 1], dtype=int32), 'three']
        >>> tree = inox.tree.mask_static(tree); tree
        [Static(1), Array([0, 1], dtype=int32), Static('three')]
        >>> jax.tree.leaves(tree)
        [Array([0, 1], dtype=int32)]
    """

    if is_static is None:
        is_static = lambda x: not is_array(x)

    return jtu.tree_map(
        f=lambda x: Static(x) if is_static(x) else x,
        tree=tree,
    )


def unmask_static(tree: PyTree) -> PyTree:
    r"""Unmasks the static leaves of a masked tree.

    See also:
        :func:`inox.tree.mask_static`

    Arguments:
        tree: The masked tree to unmask.

    Returns:
        The unmasked tree.

    Example:
        >>> tree = [Static(1), jax.numpy.arange(2), Static('three')]
        >>> inox.tree.unmask_static(tree)
        [1, Array([0, 1], dtype=int32), 'three']
    """

    is_static = lambda x: type(x) is Static

    return jtu.tree_map(
        f=lambda x: x.value if is_static(x) else x,
        tree=tree,
        is_leaf=is_static,
    )


def partition(
    tree: PyTree,
    *filters: Union[type, Callable[[Any], bool]],
    is_leaf: Callable[[Any], bool] = None,
) -> Tuple[PyTreeDef, Dict[str, Any]]:
    r"""Flattens a tree and partitions the leaves.

    The leaves are partitioned into a set of path-leaf mappings. The mapping in which a
    leaf is contained is chosen according to its oldest (closest to the root) ancestor
    that satisfies a constraint. If a node satisfies several constraints, the first one
    is selected. The last mapping is dedicated to leaves that do not satisfy any
    constraint.

    See also:
        :func:`inox.tree.combine`

    Arguments:
        tree: The tree to flatten.
        filters: A set of filtering constraints. Types are transformed into
            :py:`isinstance` constraints.
        is_leaf: A predicate for what to consider as a leaf.

    Returns:
        The tree definition and leaf partitions.

    Example:
        >>> tree = Namespace(a=1, b=jax.numpy.arange(2), c=['three', False])
        >>> treedef, leaves = inox.tree.partition(tree)
        >>> leaves
        {'.a': 1, '.b': Array([0, 1], dtype=int32), '.c[0]': 'three', '.c[1]': False}
        >>> treedef, arrays, others = inox.tree.partition(tree, jax.Array)
        >>> arrays
        {'.b': Array([0, 1], dtype=int32)}
        >>> others
        {'.a': 1, '.c[0]': 'three', '.c[1]': False}
    """

    treedef = jtu.tree_structure(tree, is_leaf)
    leaves = [{} for _ in filters] + [{}]

    def factory(filtr):
        if isinstance(filtr, type):
            return lambda x: isinstance(x, filtr)
        else:
            return filtr

    filters = list(map(factory, filters))

    if is_leaf is None:
        is_node = lambda x: any(filtr(x) for filtr in filters)
    else:
        is_node = lambda x: any(filtr(x) for filtr in filters) or is_leaf(x)

    def f(node_path, node):
        for i, filtr in enumerate(filters):  # noqa: B007
            if filtr(node):
                break
        else:
            i = len(filters)

        def g(leaf_path, leaf):
            key = jtu.keystr(node_path + leaf_path)
            leaves[i][key] = leaf

        jtu.tree_map_with_path(f=g, tree=node, is_leaf=is_leaf)

    jtu.tree_map_with_path(f=f, tree=tree, is_leaf=is_node)

    return treedef, *leaves


def combine(
    treedef: PyTreeDef,
    *leaves: Dict[str, Any],
) -> PyTree:
    r"""Reconstructs a tree from the tree definition and leaf partitions.

    See also:
        :func:`inox.tree.partition`

    Arguments:
        treedef: The tree definition.
        leaves: The set of leaf partitions.

    Returns:
        The reconstructed tree.

    Example:
        >>> tree = Namespace(a=1, b=jax.numpy.arange(2), c=['three', False])
        >>> treedef, arrays, others = inox.tree.partition(tree, jax.Array)
        >>> others = {key: str(leaf).upper() for key, leaf in others.items()}
        >>> inox.tree.combine(treedef, arrays, others)
        Namespace(
          a = '1',
          b = int32[2],
          c = ['THREE', 'FALSE']
        )
    """

    missing = []
    leaves = {key: leaf for partition in leaves for key, leaf in partition.items()}

    def f(path, leaf):
        key = jtu.keystr(path)

        if key in leaves:
            leaf = leaves.pop(key)
        else:
            missing.append(key)

        return leaf

    tree = jtu.tree_unflatten(treedef, [object()] * treedef.num_leaves)
    tree = jtu.tree_map_with_path(f, tree)

    if missing:
        keys = ", ".join(f'"{key}"' for key in missing)

        raise KeyError(f"Missing key(s) in leaves: {keys}.")

    if leaves:
        keys = ", ".join(f'"{key}"' for key in leaves)

        raise KeyError(f"Unexpected key(s) in leaves: {keys}.")

    return tree


def prepr(
    tree: PyTree,
    linewidth: int = 88,
    typeonly: bool = True,
    **kwargs,
) -> str:
    r"""Creates a pretty representation of a tree.

    Arguments:
        tree: The tree to represent.
        linewidth: The maximum line width before elements of tuples, lists and dicts
            are represented on separate lines.
        typeonly: Whether to represent the type of arrays instead of their elements.

    Returns:
        The representation string.

    Example:
        >>> tree = [1, 'two', (True, False), list(range(5)), {'6': jax.numpy.arange(7)}]
        >>> tree.append(Namespace(eight=None))
        >>> print(inox.tree.prepr(tree))
        [
          1,
          'two',
          (True, False),
          [0, 1, 2, 3, 4],
          {'6': int32[7]},
          Namespace(
            eight = None
          )
        ]
    """

    kwargs.update(
        linewidth=linewidth,
        typeonly=typeonly,
    )

    if hasattr(tree, "tree_repr"):
        return tree.tree_repr(**kwargs)
    elif dataclasses._is_dataclass_instance(tree):
        bra, ket = f"{type(tree).__name__}(", ")"
        lines = [
            f"{field.name}={prepr(getattr(tree, field.name), **kwargs)}"
            for field in dataclasses.fields(tree)
            if field.repr
        ]
    elif isinstance(tree, tuple):
        if hasattr(tree, "_fields"):
            bra, ket = f"{type(tree).__name__}(", ")"
            lines = [
                f"{field}={prepr(value, **kwargs)}" for field, value in tree._asdict().items()
            ]
        else:
            bra, ket = "(", ")"
            lines = [prepr(x, **kwargs) for x in tree]
    elif isinstance(tree, list):
        bra, ket = "[", "]"
        lines = [prepr(x, **kwargs) for x in tree]
    elif isinstance(tree, dict):
        bra, ket = "{", "}"
        lines = [f"{repr(key)}: {prepr(value, **kwargs)}" for key, value in tree.items()]
    elif is_array(tree) and typeonly:
        return f"{tree.dtype}{list(tree.shape)}"
    else:
        return repr(tree).strip(" \n")

    if any("\n" in line for line in lines):
        lines = ",\n".join(lines)
    elif sum(len(line) + 2 for line in lines) > linewidth:
        lines = ",\n".join(lines)
    else:
        lines = ", ".join(lines)

    if "\n" in lines:
        lines = "\n" + indent(lines, "  ") + "\n"

    return f"{bra}{lines}{ket}"



r"""Sharing modules

In a vanilla :class:`inox.nn.module.Module`, shared references to the same layer or
parameter would be treaded as separate copies and their weights would not be tied. The
:class:`Scope` class correctly handles such cases when shared references are explicited
with :class:`Reference`.

.. code-block:: python

    import inox
    import inox.nn
    import jax
    import jax.numpy as jnp

    class WeightSharingMLP(nn.Scope):
        def __init__(self, key):
            keys = jax.random.split(key, 3)

            self.l1 = nn.Linear(in_features=64, out_features=64, key=keys[0])
            self.l3 = nn.Linear(in_features=64, out_features=64, key=keys[1])
            self.l4 = nn.Linear(in_features=64, out_features=64, key=keys[2])

            self.l1 = nn.Reference('l1', self.l1)
            self.l2 = self.l1  # tied layer
            self.l3.weight = nn.Reference('l3.weight', self.l3.weight)
            self.l4.weight = self.l3.weight  # tied parameter

            self.relu = nn.ReLU()

        def __call__(self, x):  # standard __call__
            x = self.l1(x)
            x = self.l2(self.relu(x))
            x = self.l3(self.relu(x))
            x = self.l4(self.relu(x))

            return x

    key = jax.random.key(0)
    model = WeightSharingMLP(key)
    static, params, others = model.partition(nn.Parameter)

    print(inox.tree_repr(params))  # does not contain 'l2' and 'l4.weight'

.. code-block:: text

    {
      '.l1.value.bias.value': float32[64],
      '.l1.value.weight.value': float32[64, 64],
      '.l3.bias.value': float32[64],
      '.l3.weight.value.value': float32[64, 64],
      '.l4.bias.value': float32[64]
    }
"""

__all__ = [
    "Scope",
    "Reference",
]

import jax.tree_util as jtu

from typing import Any, Hashable

import inox.tree

from .module import Module


class Scope(Module):
    r"""Subclass of :class:`inox.nn.module.Module` which handles shared object
    references within its scope.

    All references with the same identification tag in a scope are considered to be the
    same and all but one copies are pruned during the flattening of the scope tree.
    Cyclic references are allowed, with the exception of a scope referencing itself.

    Warning:
        Shared references and in-place mutations are very hard to combine properly.
        Conversely, :class:`Reference` works seamlessly with :mod:`inox.nn.state` utils.

    Arguments:
        kwargs: A name-value mapping.
    """

    def tree_flatten(self):
        values, names = super().tree_flatten()
        visited = set()

        def prune(tree):
            def f(x):
                if isinstance(x, Reference):
                    if x.tag in visited:
                        x = Reference(x.tag, None)
                    else:
                        visited.add(x.tag)
                        x = Reference(x.tag, prune(x.value))

                return x

            def is_leaf(x):
                return isinstance(x, Reference) or isinstance(x, Scope)

            return jtu.tree_map(f, tree, is_leaf=is_leaf)

        return prune(values), names

    @classmethod
    def tree_unflatten(cls, names, values):
        visited = {}

        def unprune(tree):
            def f(x):
                if isinstance(x, Reference):
                    if x.tag in visited:
                        x = visited[x.tag]
                    else:
                        visited[x.tag] = x
                        x = Reference(x.tag, unprune(x.value))

                return x

            def is_leaf(x):
                return isinstance(x, Reference) or isinstance(x, Scope)

            return jtu.tree_map(f, tree, is_leaf=is_leaf)

        return super().tree_unflatten(names, unprune(values))

    def tree_repr(self, **kwargs) -> str:
        kwargs["references"] = set()
        return super().tree_repr(**kwargs)


class Reference(metaclass=inox.tree.PyTreeMeta):
    r"""Creates a reference to a value.

    A :class:`Reference` instance forwards :py:`__call__`,  :py:`__getattr__`, and
    :py:`__getitem__` operations to the value it references. For arithmetic operations
    (`+`, `*`, ...), use :py:`ref.value` directly instead.

    See also:
        :class:`Scope`

    Arguments:
        tag: An identification tag.
        value: The value to reference.

    Example:
        >>> weight = Reference('my-ref', nn.Parameter(jax.numpy.ones((3, 5))))
        >>> weight  # repr preceded by an asterisk
        *Parameter(float32[3, 5])
        >>> weight.shape
        (3, 5)
        >>> weight()
        Array([[1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1.],
               [1., 1., 1., 1., 1.]], dtype=float32)
    """

    value: Any = None

    def __init__(self, tag: Hashable, value: Any):
        self.tag = tag
        self.value = value

    def __call__(self, *args, **kwargs) -> Any:
        return self.value(*args, **kwargs)

    def __delattr__(self, name: str):
        if name in ("tag", "value"):
            object.__delattr__(self, name)
        else:
            delattr(self.value, name)

    def __delitem__(self, key: Hashable):
        del self.value[key]

    def __getattr__(self, name: str) -> Any:
        return getattr(self.value, name)

    def __getitem__(self, key: Hashable):
        return self.value[key]

    def __setattr__(self, name: str, value: Any):
        if name in ("tag", "value"):
            object.__setattr__(self, name, value)
        else:
            setattr(self.value, name, value)

    def __setitem__(self, key: Hashable, value: Any):
        self.value[key] = value

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        references = kwargs.setdefault("references", set())

        if self.tag in references:
            return f"@{self.tag}"
        else:
            references.add(self.tag)
            return f"*{inox.tree.prepr(self.value, **kwargs)}"

    def tree_flatten(self):
        return [self.value], self.tag

    def tree_flatten_with_keys(self):
        return [(jtu.GetAttrKey("value"), self.value)], self.tag

    @classmethod
    def tree_unflatten(cls, tag, children):
        self = object.__new__(cls)
        self.tag = tag
        self.value = children[0]

        return self



r"""Tests for the inox.tree_util module."""

import jax.numpy as jnp
import jax.tree_util as jtu
import pickle
import pytest

from jax import Array
from typing import Hashable

from inox.tree import Namespace, Static, combine, mask_static, partition, unmask_static


def tree_eq(x, y):
    def eq(a, b):
        if isinstance(a, Array) and isinstance(b, Array):
            return jnp.allclose(a, b)
        elif isinstance(a, Array) or isinstance(b, Array):
            return False
        else:
            return a == b

    return jtu.tree_all(jtu.tree_map(eq, x, y))


def test_Namespace():
    x = Namespace(a=1, b="2")

    assert x.a == 1
    assert x.b == "2"

    # Add
    x.c = True

    assert x.c is True

    # Modify
    x.a = Namespace(d=None)

    assert x.a.d is None

    # Delete
    del x.b

    assert not hasattr(x, "b")

    # Flatten
    leaves, treedef = jtu.tree_flatten(x)

    assert isinstance(treedef, Hashable)

    y = jtu.tree_unflatten(treedef, leaves)

    assert tree_eq(x, y)

    # Pickle
    data = pickle.dumps(x)
    y = pickle.loads(data)

    assert tree_eq(x, y)

    # Print
    assert repr(x)


@pytest.mark.parametrize("nested", [False, True])
def test_Static(nested: bool):
    if nested:
        x = Static(Static("hashable"))
    else:
        x = Static("hashable")

    # Flatten
    leaves, treedef = jtu.tree_flatten(x)

    assert not leaves
    assert isinstance(treedef, Hashable)

    y = jtu.tree_unflatten(treedef, leaves)

    assert x == y

    # Pickle
    data = pickle.dumps(x)
    y = pickle.loads(data)

    assert x == y

    # Print
    assert repr(x)


def test_mask_static():
    x = Namespace(
        a=jnp.ones(1),
        b=2,
        c=[jnp.arange(3), False],
        d=Namespace(e="five", f=jnp.eye(6)),
    )

    # mask_static
    y = mask_static(x)

    leaves, treedef = jtu.tree_flatten(y)

    assert all(isinstance(leaf, Array) for leaf in leaves)
    assert isinstance(treedef, Hashable)

    # unmask_static
    z = unmask_static(y)

    leaves, treedef = jtu.tree_flatten(z)

    assert not all(isinstance(leaf, Array) for leaf in leaves)
    assert isinstance(treedef, Hashable)

    assert tree_eq(x, z)


def test_partition():
    x = Namespace(
        a=jnp.ones(1),
        b=2,
        c=[jnp.arange(3), False],
        d=Namespace(e="five", f=jnp.eye(6)),
    )

    # partition
    treedef_a, leaves = partition(x)

    assert not all(isinstance(leaf, Array) for leaf in leaves.values())
    assert isinstance(treedef_a, Hashable)

    treedef_b, arrays, others = partition(x, Array)

    assert all(isinstance(leaf, Array) for leaf in arrays.values())
    assert not any(isinstance(leaf, Array) for leaf in others.values())
    assert isinstance(treedef_b, Hashable)

    assert treedef_a == treedef_b

    # combine
    y = combine(treedef_a, leaves)
    z = combine(treedef_b, arrays, others)

    assert tree_eq(x, y)
    assert tree_eq(x, z)

    with pytest.raises(KeyError):
        combine(treedef_a, {})

    with pytest.raises(KeyError):
        combine(treedef_a, leaves, {"none": None})



r"""Tests for the inox.api module."""

from functools import partial

from inox.api import automask, inner, outer
from inox.tree import Static


class SubStatic(Static):
    pass


def test_automask():
    def is_static(x):
        assert type(x) is Static
        return x

    def is_not_static(x):
        assert type(x) is not Static
        return x

    def is_sub_static(x):
        assert type(x) is SubStatic
        return x

    # inner & outer
    assert is_static(inner(is_not_static)("leaf"))
    assert is_static(inner(is_not_static)(Static("leaf")))
    assert is_sub_static(inner(is_sub_static)(SubStatic("leaf")))
    assert is_not_static(outer(is_static)("leaf"))
    assert is_not_static(outer(is_static)(Static("leaf")))
    assert is_sub_static(outer(is_sub_static)(SubStatic("leaf")))

    def f(x):
        return x

    assert outer(inner(f)) is f
    assert inner(outer(f)) is f

    # automask
    assert is_not_static(automask(partial)(is_not_static)("leaf"))
    assert is_not_static(automask(partial)(is_not_static)(Static("leaf")))
    assert is_sub_static(automask(partial)(is_sub_static)(SubStatic("leaf")))

    assert automask(lambda f: f)(f) is f
    assert automask(partial)(f) is not f



r"""Extended user-facing transformations and utilities.

The transformations provided in the :mod:`inox.api` module are lifted versions of native
JAX transformations for which all non-array leaves (:py:`float`, :py:`str`, functions,
...) are considered static, that is part of the tree structure.
"""

__all__ = [
    "automask",
    "jit",
    "grad",
    "value_and_grad",
    "jacfwd",
    "jacrev",
    "hessian",
    "checkpoint",
    "vmap",
    "pmap",
]

import jax

from functools import cache, wraps
from jax._src.api import api_boundary
from typing import Callable

from inox.tree import mask_static, unmask_static


@cache
def inner(fun: Callable):
    if fun is getattr(fun, "__inner__", None):
        return fun

    if fun is getattr(fun, "__outer__", None):
        return fun.__wrapped__

    @wraps(fun)
    @api_boundary
    def wrapped(*args, **kwargs):
        return mask_static(fun(*unmask_static(args), **unmask_static(kwargs)))

    wrapped.__inner__ = wrapped

    return wrapped


def outer(fun: Callable):
    if fun is getattr(fun, "__inner__", None):
        return fun.__wrapped__

    if fun is getattr(fun, "__outer__", None):
        return fun

    @wraps(fun)
    @api_boundary
    def wrapped(*args, **kwargs):
        return unmask_static(fun(*mask_static(args), **mask_static(kwargs)))

    wrapped.__outer__ = wrapped

    return wrapped


def automask(transform: Callable) -> Callable:
    r"""Lifts a transformation to consider all non-array leaves as static.

    For a function :py:`f` and a JAX transformation :py:`jax.tf`,

    .. code-block:: python

        y = automask(jax.tf)(f)(x)

    is equivalent to

    .. code-block:: python

        g = lambda x: inox.tree.mask_static(f(inox.tree.unmask_static(x)))
        y = inox.tree.unmask_static(jax.tf(g)(inox.tree.mask_static(x)))

    See also:
        :func:`inox.tree.mask_static`

    Arguments:
        transform: The transformation to lift.

    Returns:
        The lifted transformation.
    """

    @wraps(transform)
    @api_boundary
    def wrapped(fun: Callable, *args, **kwargs) -> Callable:
        return outer(transform(inner(fun), *args, **kwargs))

    return wrapped


jit = automask(jax.jit)
grad = automask(jax.grad)
value_and_grad = automask(jax.value_and_grad)
jacfwd = automask(jax.jacfwd)
jacrev = automask(jax.jacrev)
hessian = automask(jax.hessian)
checkpoint = automask(jax.checkpoint)
vmap = automask(jax.vmap)
pmap = automask(jax.pmap)



r"""Pooling layers"""

__all__ = [
    "AvgPool",
    "MaxPool",
]

import jax
import math

from functools import wraps
from jax import Array
from typing import Callable, Sequence, Tuple, Union

from .module import Module
from ..numpy import vectorize


class Pool(Module):
    r"""Abstract spatial pooling class."""

    def __init__(
        self,
        window_size: Sequence[int],
        stride: Union[int, Sequence[int]] = None,
        padding: Union[int, Sequence[Tuple[int, int]]] = 0,
    ):
        if stride is None:
            stride = window_size
        elif isinstance(stride, int):
            stride = [stride] * len(window_size)

        if isinstance(padding, int):
            padding = [(padding, padding)] * len(window_size)

        self.window_size = window_size
        self.stride = stride
        self.padding = padding

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C)`,
            such that

            .. math:: H_i' = \left\lfloor \frac{H_i - k_i + p_i}{s_i} + 1 \right\rfloor

            where :math:`k_i`, :math:`s_i` and :math:`p_i` are respectively the window
            size, the stride coefficient and the total padding of the :math:`i`-th
            spatial axis.
        """

        return vectorize(jax.lax.reduce_window, ndims=self.ndim)(
            x,
            init_value=self.initial,
            computation=self.operator,
            window_dimensions=(*self.window_size, 1),
            window_strides=(*self.stride, 1),
            padding=(*self.padding, (0, 0)),
        )

    @property
    def ndim(self) -> int:
        return len(self.window_size) + 1


class AvgPool(Pool):
    r"""Creates an average spatial pooling layer.

    Arguments:
        window_size: The size of the pooling window in each spatial axis.
        stride: The stride coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
    """

    @wraps(Pool.__call__)
    def __call__(self, x: Array) -> Array:
        return super().__call__(x) / math.prod(self.window_size)

    @property
    def operator(self) -> Callable[[Array, Array], Array]:
        return jax.lax.add

    @property
    def initial(self) -> float:
        return 0.0


class MaxPool(Pool):
    r"""Creates a maximum spatial pooling layer.

    Arguments:
        window_size: The size of the pooling window in each spatial axis.
        stride: The stride coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
    """

    @wraps(Pool.__call__)
    def __call__(self, x: Array) -> Array:
        return super().__call__(x)

    @property
    def operator(self) -> Callable[[Array, Array], Array]:
        return jax.lax.max

    @property
    def initial(self) -> float:
        return -math.inf



r"""Einstein-like tensor operations

References:
    | Einops: Clear and Reliable Tensor Manipulations with Einstein-like Notation (Rogozhnikov et al., 2022)
    | https://openreview.net/forum?id=oapKSVM2bcj
"""

__all__ = [
    "Rearrange",
    "Reduce",
    "Repeat",
]

import einops

from jax import Array

from .module import Module


class Rearrange(Module):
    r"""Creates an axis rearrangement layer.

    This module is a thin wrapper around :func:`einops.rearrange`.

    Arguments:
        pattern: The axis rearrangement pattern. For example, the pattern
            :py:`'A B C -> C (A B)'` moves and flattens the two first axes.
        lengths: The lengths of the axes.
    """

    def __init__(self, pattern: str, **lengths: int):
        self.pattern = pattern
        self.lengths = lengths

    def __call__(self, x: Array) -> Array:
        return einops.rearrange(x, self.pattern, **self.lengths)


class Reduce(Module):
    r"""Creates an axis reduction layer.

    This module is a thin wrapper around :func:`einops.reduce`.

    Arguments:
        pattern: The axis rearrangement pattern. For example, the pattern
            :py:`'A B C -> A C'` reduces the second axis.
        reduction: The type of reduction (:py:`'sum'`, :py:`'mean'`, :py:`'max'`, ...).
        lengths: The lengths of the axes.
    """

    def __init__(self, pattern: str, reduction: str = "sum", **lengths: int):
        self.pattern = pattern
        self.reduction = reduction
        self.lengths = lengths

    def __call__(self, x: Array) -> Array:
        return einops.reduce(x, self.pattern, self.reduction, **self.lengths)


class Repeat(Module):
    r"""Creates an axis repetition layer.

    This module is a thin wrapper around :func:`einops.repeat`.

    Arguments:
        pattern: The axis rearrangement pattern. For example, the pattern
            :py:`'A B -> A C B'` inserts a new axis.
        lengths: The lengths of the axes.
    """

    def __init__(self, pattern: str, **lengths: int):
        self.pattern = pattern
        self.lengths = lengths

    def __call__(self, x: Array) -> Array:
        return einops.repeat(x, self.pattern, **self.lengths)



r"""Dropout layers"""

__all__ = [
    "Dropout",
    "TrainingDropout",
]

import jax
import jax.numpy as jnp

from jax import Array
from typing import Union

from .module import Module
from ..random import get_rng


class Dropout(Module):
    r"""Creates a dropout layer.

    .. math:: y = \frac{m \odot x}{1 - p}

    where the binary mask :math:`m` is drawn from a Bernoulli distribution such that
    :math:`P(m_i = 0) = p`. This has proven to be an effective technique for
    regularization and preventing overfitting.

    References:
        | A Simple Way to Prevent Neural Networks from Overfitting (Srivastava et al., 2014)
        | https://jmlr.org/papers/v15/srivastava14a.html

    Arguments:
        p: The dropout rate :math:`p \in [0, 1]`.
    """

    def __init__(self, p: Union[float, Array] = 0.5):
        self.p = jnp.asarray(p)

    def __call__(self, x: Array, key: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*)`.
            key: A PRNG key.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*)`.
        """

        mask = jax.random.bernoulli(key, 1 - self.p, shape=x.shape)

        return jnp.where(mask, x / (1 - self.p), 0)


class TrainingDropout(Dropout):
    r"""Creates a training-bound dropout layer.

    When :py:`self.training = False`,

    .. math:: y = x

    See also:
        :class:`Dropout`

    Arguments:
        p: The dropout rate :math:`p \in [0, 1]`.
    """

    training: bool = True

    def __call__(self, x: Array, key: Array = None) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*)`.
            key: A PRNG key. If :py:`None`, :py:`inox.random.get_rng("dropout")`
                is used instead.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*)`.
        """

        if self.training:
            if key is None:
                key = get_rng("dropout").split()

            return super().__call__(x, key)
        else:
            return x



r"""Tests for the inox.random module."""

import jax
import jax.numpy as jnp
import pickle
import pytest

from typing import Hashable

from inox.random import PRNG, get_rng, set_rng


@pytest.mark.parametrize("seed", [0, jax.random.key(0)])
def test_PRNG(seed):
    rng = PRNG(seed)
    shape = rng.state.shape

    # Split
    assert rng.split().shape == shape
    assert rng.split(3).shape == (3, *shape)

    assert not jnp.allclose(rng.split(), rng.split())

    # Sample
    assert rng.normal().shape == ()
    assert rng.uniform(shape=(2, 3)).shape == (2, 3)

    assert not jnp.allclose(rng.normal(), rng.normal())

    # Detect JIT
    with pytest.raises(AssertionError):
        jax.jit(lambda: rng.split())()

    # Flatten
    leaves, treedef = jax.tree_util.tree_flatten(rng)

    assert isinstance(treedef, Hashable)

    jax.tree_util.tree_unflatten(treedef, leaves)

    # Pickle
    data = pickle.dumps(rng)
    pickle.loads(data)

    # Print
    assert repr(rng)


@pytest.mark.parametrize("seed", [0, jax.random.key(0)])
def test_set_rng(seed):
    with set_rng(init=PRNG(seed), dropout=PRNG(seed)):
        a = get_rng("init").split()
        b = get_rng("dropout").split()

        assert jnp.allclose(a, b)

    with pytest.raises(AssertionError):
        get_rng("init")



r"""Attention layers"""

__all__ = [
    "MultiheadAttention",
]

import jax
import jax.numpy as jnp
import math

from einops import rearrange
from jax import Array
from typing import Union

from .linear import Linear
from .module import Module
from ..random import get_rng


def attention(
    q: Array,
    k: Array,
    v: Array,
    mask: Array = None,
) -> Array:
    r"""Computes the scaled dot-product attention.

    Arguments:
        q: The query tensor :math:`Q`, with shape :math:`(*, S, C)`.
        k: The key tensor :math:`K`, with shape :math:`(*, T, C)`.
        v: The value tensor :math:`V`, with shape :math:`(*, T, C')`.
        mask: A boolean attention mask, with shape :math:`(*, S, T)`.
            A :py:`False` value indicates that the corresponding attention weight
            is set to :math:`-\infty`.

    Returns:
        The output vector :math:`y`, with shape :math:`(*, S, C')`.
    """

    C = q.shape[-1]

    weight = jnp.einsum("...ik,...jk->...ij", q, k)
    weight = weight / math.sqrt(C)

    if mask is not None:
        weight = jnp.where(mask, weight, -1e9)

    attn = jax.nn.softmax(weight, axis=-1)

    return jnp.einsum("...ij,...jk->...ik", attn, v)


class MultiheadAttention(Module):
    r"""Creates a multihead attention layer.

    .. math:: Y = \sum_i
        \mathrm{attention}(X_q W_q^i + b_q^i, X_k W_k^i + b_k^i, X_v W_v^i + b_v^i) W_y^i

    where

    .. math:: \mathrm{attention}(Q, K, V) =
        \mathrm{softmax}\left( \frac{Q K^T}{\sqrt{H}} \right) V

    denotes the scaled dot-product attention.

    References:
        | Attention Is All You Need (Vaswani et al., 2023)
        | https://arxiv.org/abs/1706.03762

    Arguments:
        heads: The number of attention heads :math:`N`.
        in_features: The number of input features :math:`C`.
        out_features: The number of output features :math:`C'`.
            If :py:`None`, :math:`C' = C`.
        hid_features: The number of hidden features :math:`H` per head.
            If :py:`None`, :math:`H = \frac{C}{N}`.
        bias: Whether the layer learns additive biases :math:`(b_q, b_k, b_v)` or not.
        causal: Whether the attention mask is causal or not. If :py:`True`, the
            :math:`i`-th query is only allowed to attend the :math:`j`-th key if
            :math:`j - i \leq T - S`.
        dropout: The dropout rate on attention weights.
        key: A PRNG key for initialization. If :py:`None`,
            :py:`inox.random.get_rng("init")` is used instead.
    """

    def __init__(
        self,
        heads: int,
        in_features: int,
        out_features: int = None,
        hid_features: int = None,
        bias: bool = True,
        causal: bool = False,
        dropout: Union[float, Array] = 0.0,
        key: Array = None,
    ):
        if key is None:
            keys = get_rng("init").split(4)
        else:
            keys = jax.random.split(key, 4)

        if out_features is None:
            out_features = in_features

        if hid_features is None:
            hid_features = in_features // heads

        self.lin_q = Linear(in_features, hid_features * heads, bias, key=keys[0])
        self.lin_k = Linear(in_features, hid_features * heads, bias, key=keys[1])
        self.lin_v = Linear(in_features, hid_features * heads, bias, key=keys[2])
        self.lin_y = Linear(hid_features * heads, out_features, False, key=keys[3])

        self.heads = heads
        self.causal = causal
        self.dropout = jnp.asarray(dropout)

    def __call__(
        self,
        xq: Array,
        xk: Array = None,
        xv: Array = None,
        mask: Array = None,
        key: Array = None,
    ) -> Array:
        r"""
        Arguments:
            xq: The query tensor :math:`X_q`, with shape :math:`(*, S, C)`.
            xk: The key tensor :math:`X_k`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_k = X_q`.
            xv: The value tensor :math:`X_v`, with shape :math:`(*, T, C)`.
                If :py:`None`, :math:`X_v = X_k`.
            mask: A boolean attention mask, with shape :math:`(*, S, T)`.
                A :py:`False` value indicates that the corresponding attention weight
                is set to :math:`-\infty`.
            key: A PRNG key. If :py:`None`, dropout is not applied.

        Returns:
            The output tensor :math:`Y`, with shape :math:`(*, S, C')`.
        """

        if xk is None:
            xk = xq

        if xv is None:
            xv = xk

        S, T = xq.shape[-2], xk.shape[-2]

        # Project
        q = self.lin_q(xq)
        k = self.lin_k(xk)
        v = self.lin_v(xv)

        q, k, v = [rearrange(x, "... L (N H) -> ... N L H", N=self.heads) for x in (q, k, v)]

        # Mask
        if self.causal:
            if mask is None:
                mask = jnp.ones((S, T), dtype=bool)

            mask = jnp.tril(mask, T - S)

        if key is not None:
            shape = jnp.broadcast_shapes(
                (*q.shape[:-2], S, 1),
                (*k.shape[:-2], 1, T),
                (S, T) if mask is None else mask.shape,
            )

            keep = jax.random.bernoulli(key, p=1 - self.dropout, shape=shape)

            if mask is None:
                mask = keep
            else:
                mask = jnp.logical_and(mask, keep)

        # Attention
        y = attention(q, k, v, mask)
        y = rearrange(y, "... N L H -> ... L (N H)")
        y = self.lin_y(y)

        return y



r"""State space model (SSM) layers"""

__all__ = [
    "SISO",
    "S4",
]

import jax
import jax.numpy as jnp
import math
import numpy as np

from jax import Array
from typing import Tuple

from .module import ComplexParameter, Module, Parameter
from ..random import get_rng


class SISO(Module):
    r"""Abstract single-input single-output (SISO) state space model class.

    A SISO state space model defines a system of equations of the form

    .. math::
        \dot{x}(t) & = A x(t) + B u(t) \\
        y(t) & = C x(t)

    where :math:`u(t), y(t) \in \mathbb{C}` are input and output signals and :math:`x(t)
    \in \mathbb{C}^{H}` is a latent/hidden state. In practice, the input and output
    signals are sampled every :math:`\Delta` time units, leading to sequences
    :math:`(x_1, x_2, \dots)` and :math:`(y_1, y_2, \dots)` whose dynamics are governed
    by the discrete-time form of the system

    .. math::
        x_i & = \bar{A} x_{i-1} + \bar{B} u_i \\
        y_i & = \bar{C} x_i

    where :math:`\bar{A} = \exp(\Delta A)`, :math:`\bar{B} = A^{-1} (\bar{A} - I) B` and
    :math:`\bar{C} = C`. Assuming :math:`x_0 = 0`, the dynamics can also be represented
    as a discrete-time convolution

    .. math:: y_{1:L} = \bar{k}_{1:L} * u_{1:L}

    where :math:`\bar{k}_i = \bar{C} \bar{A}^{i-1} \bar{B} \in \mathbb{C}`.

    Wikipedia:
        https://wikipedia.org/wiki/State-space_representation
    """

    def __call__(self, u: Array) -> Array:
        r"""
        Arguments:
            u: The input signal :math:`u_{1:L}`, with shape :math:`(*, L)`.
                Floating point arrays are promoted to complex arrays.

        Returns:
            The output signal :math:`y_{1:L}`, with shape :math:`(*, L)`.
        """

        L = u.shape[-1]
        k = self.kernel(L)

        return jax.scipy.signal.fftconvolve(k, u, axes=-1)[..., :L]

    def discrete(self) -> Tuple[Array, Array, Array]:
        r"""
        Returns:
            The matrices :math:`\bar{A}`, :math:`\bar{B}` and :math:`\bar{C}`,
            respectively with shape :math:`(H, H)`, :math:`(H)` and :math:`(H)`.
        """

        raise NotImplementedError()

    def kernel(self, length: int) -> Array:
        r"""
        Arguments:
            length: The kernel length :math:`L`.

        Returns:
            The kernel :math:`\bar{k}_{1:L}`, with shape :math:`(L)`.
        """

        raise NotImplementedError()


class Eigenvalue(ComplexParameter):
    r"""Wrapper to indicate an optimizable eigenvalue array."""

    def __init__(self, value: Array):
        self.log_real = jnp.log(-value.real)
        self.imag = value.imag

    @property
    def real(self) -> Array:
        return -jnp.exp(self.log_real)


class S4(SISO):
    r"""Creates an S4 state space model.

    References:
        | Efficiently Modeling Long Sequences with Structured State Spaces (Gu et al., 2021)
        | https://arxiv.org/abs/2111.00396

        | The Annotated S4 (Rush et al., 2023)
        | https://srush.github.io/annotated-s4

    Arguments:
        hid_features: The number of hidden features :math:`H`.
        key: A PRNG key for initialization. If :py:`None`,
            :func:`inox.random.get_rng` is used instead.

    Example:
        >>> ssm = S4(hid_features=64, key=jax.random.key(0))
        >>> u = jax.numpy.linspace(0.0, 1.0, 1024)
        >>> y = ssm(u)
    """

    def __init__(self, hid_features: int, key: Array = None):
        if key is None:
            keys = get_rng("init").split(3)
        else:
            keys = jax.random.split(key, 3)

        A, P = S4.DPLR_HiPPO(hid_features)

        self.A = Eigenvalue(A)
        self.P = ComplexParameter(P)
        self.B = ComplexParameter(
            jax.random.normal(
                keys[0],
                shape=(hid_features,),
                dtype=complex,
            )
        )
        self.C = ComplexParameter(
            jax.random.normal(
                keys[1],
                shape=(hid_features,),
                dtype=complex,
            )
            / math.sqrt(hid_features)
        )
        self.log_dt = Parameter(
            jax.random.uniform(
                keys[2],
                shape=(),
                minval=math.log(1e-3),
                maxval=math.log(1e-1),
            )
        )

    @staticmethod
    def DPLR_HiPPO(n: int) -> Tuple[Array, Array]:
        r"""Returns the diagonal plus low-rank (DPLR) form of the HiPPO matrix.

        .. math:: A = \Lambda - PP^*

        Arguments:
            n: The size :math:`n` of the HiPPO matrix.
        """

        P = np.sqrt(np.arange(n) + 1 / 2)
        S = np.outer(P, P.conj())
        S = np.tril(S) - np.triu(S)

        # Diagonal A
        A_real = -np.ones(n) / 2
        A_imag, V = np.linalg.eigh(-1j * S)
        A = A_real + 1j * A_imag

        # Project P
        P = V.T.conj() @ P

        return A, P

    def discrete(self) -> Tuple[Array, Array, Array]:
        A, P, B, C = self.A(), self.P(), self.B(), self.C()
        dt = jnp.exp(self.log_dt())

        D = jnp.diag(1 / (2 / dt - A))
        PQ = jnp.outer(P, P.conj())
        A0 = jnp.diag(2 / dt + A) - PQ
        A1 = D - D @ PQ / (1 + jnp.dot(P.conj(), D @ P)) @ D

        Ab = A1 @ A0
        Bb = 2 * A1 @ B

        return Ab, Bb, C

    def kernel(self, length: int) -> Array:
        A, P, B, C = self.A(), self.P(), self.B(), self.C()
        dt = jnp.exp(self.log_dt())

        # \tilde{C}
        Ab, _, _ = self.discrete()
        Ct = C - jnp.linalg.matrix_power(Ab, length).T @ C

        # Roots of unity
        z = jnp.exp(-2j * math.pi / length * jnp.arange(length))

        # Cauchy
        w = 2 / dt * (1 - z) / (1 + z)
        v = jnp.stack((
            Ct * B,
            Ct * P,
            P.conj() * B,
            P.conj() * P,
        ))

        k00, k01, k10, k11 = jnp.sum(v[:, None] / (w[:, None] - A), axis=-1)

        # Kernel
        k = 2 / (1 + z) * (k00 - k01 / (1 + k11) * k10)
        k = jnp.fft.ifft(k)

        return k



r"""Linear layers"""

__all__ = [
    "Linear",
    "Conv",
    "ConvTransposed",
]

import jax
import jax.numpy as jnp
import math

from jax import Array
from typing import Sequence, Tuple, Union

from .module import Module, Parameter
from ..numpy import vectorize
from ..random import get_rng


class Linear(Module):
    r"""Creates a linear layer.

    .. math:: y = W x + b

    Arguments:
        in_features: The number of input features :math:`C`.
        out_features: The number of output features :math:`C'`.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        key: A PRNG key for initialization. If :py:`None`,
            :py:`inox.random.get_rng("init")` is used instead.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        key: Array = None,
    ):
        if key is None:
            key = get_rng("init").split()

        scale = 1 / math.sqrt(in_features)

        self.weight = Parameter(
            jax.random.uniform(
                key,
                shape=(in_features, out_features),
                minval=-scale,
                maxval=scale,
            )
        )

        if bias:
            self.bias = Parameter(jnp.zeros(out_features))
        else:
            self.bias = None

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input vector :math:`x`, with shape :math:`(*, C)`.

        Returns:
            The output vector :math:`y`, with shape :math:`(*, C')`.
        """

        if self.bias is None:
            return x @ self.weight()
        else:
            return x @ self.weight() + self.bias()


class Conv(Module):
    r"""Creates a convolution layer.

    .. math:: y = W * x + b

    References:
        | A guide to convolution arithmetic for deep learning (Dumoulin et al., 2016)
        | https://arxiv.org/abs/1603.07285

    Arguments:
        in_channels: The number of input channels :math:`C`.
        out_channels: The number of output channels :math:`C'`.
        kernel_size: The size of the kernel :math:`W` in each spatial axis.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        stride: The stride coefficient in each spatial axis.
        dilation: The dilation coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
        groups: The number of channel groups :math:`G`.
            Both :math:`C` and :math:`C'` must be divisible by :math:`G`.
        key: A PRNG key for initialization. If :py:`None`,
            :func:`inox.random.get_rng("init")` is used instead.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Sequence[int],
        bias: bool = True,
        stride: Union[int, Sequence[int]] = 1,
        dilation: Union[int, Sequence[int]] = 1,
        padding: Union[int, Sequence[Tuple[int, int]]] = 0,
        groups: int = 1,
        key: Array = None,
    ):
        if key is None:
            key = get_rng("init").split()

        if isinstance(stride, int):
            stride = [stride] * len(kernel_size)

        if isinstance(dilation, int):
            dilation = [dilation] * len(kernel_size)

        if isinstance(padding, int):
            padding = [(padding, padding)] * len(kernel_size)

        in_channels = in_channels // groups
        scale = 1 / math.sqrt(math.prod(kernel_size) * in_channels)

        self.kernel = Parameter(
            jax.random.uniform(
                key,
                shape=(*kernel_size, in_channels, out_channels),
                minval=-scale,
                maxval=scale,
            )
        )

        if bias:
            self.bias = Parameter(jnp.zeros(out_channels))
        else:
            self.bias = None

        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.padding = padding
        self.groups = groups

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C')`,
            such that

            .. math:: H_i' =
                \left\lfloor \frac{H_i - d_i \times (k_i - 1) + p_i}{s_i} + 1 \right\rfloor

            where :math:`k_i`, :math:`s_i`, :math:`d_i` and :math:`p_i` are respectively
            the kernel size, the stride coefficient, the dilation coefficient and the
            total padding of the :math:`i`-th spatial axis.
        """

        x = jnp.expand_dims(x, axis=0)
        x = vectorize(jax.lax.conv_general_dilated, ndims=self.ndim + 1)(
            x,
            rhs=self.kernel(),
            dimension_numbers=self.dimensions,
            window_strides=self.stride,
            rhs_dilation=self.dilation,
            padding=self.padding,
            feature_group_count=self.groups,
        )
        x = jnp.squeeze(x, axis=0)

        if self.bias is None:
            return x
        else:
            return x + self.bias()

    @property
    def ndim(self) -> int:
        return len(self.kernel_size) + 1

    @property
    def dimensions(self) -> jax.lax.ConvDimensionNumbers:
        return jax.lax.ConvDimensionNumbers(
            (0, self.ndim, *range(1, self.ndim)),
            (self.ndim, self.ndim - 1, *range(0, self.ndim - 1)),
            (0, self.ndim, *range(1, self.ndim)),
        )


class ConvTransposed(Conv):
    r"""Creates a transposed convolution layer.

    This layer can be seen as the gradient of :class:`Conv` with respect to its input.
    It is also known as a "deconvolution", altough it does not actually compute the
    inverse of a convolution.

    References:
        | A guide to convolution arithmetic for deep learning (Dumoulin et al., 2016)
        | https://arxiv.org/abs/1603.07285

    Arguments:
        in_channels: The number of input channels :math:`C`.
        out_channels: The number of output channels  :math:`C'`.
        kernel_size: The size of the kernel :math:`W` in each spatial axis.
        bias: Whether the layer learns an additive bias :math:`b` or not.
        stride: The stride coefficient in each spatial axis.
        dilation: The dilation coefficient in each spatial axis.
        padding: The padding applied to each end of each spatial axis.
        groups: The number of channel groups :math:`G`.
            Both :math:`C` and :math:`C'` must be divisible by :math:`G`.
        key: A PRNG key for initialization. If :py:`None`,
            :func:`inox.random.get_rng` is used instead.
    """

    def __call__(self, x: Array) -> Array:
        r"""
        Arguments:
            x: The input tensor :math:`x`, with shape :math:`(*, H_1, \dots, H_n, C)`.

        Returns:
            The output tensor :math:`y`, with shape :math:`(*, H_1', \dots, H_n', C')`,
            such that

            .. math:: H_i' = (H_i - 1) \times s_i + d_i \times (k_i - 1) - p_i + 1

            where :math:`k_i`, :math:`s_i`, :math:`d_i` and :math:`p_i` are respectively
            the kernel size, the stride coefficient, the dilation coefficient and the
            total padding of the :math:`i`-th spatial axis.
        """

        x = jnp.expand_dims(x, axis=0)
        x = vectorize(jax.lax.conv_general_dilated, ndims=self.ndim + 1)(
            x,
            rhs=self.kernel(),
            dimension_numbers=self.dimensions,
            window_strides=[1] * (self.ndim - 1),
            padding=self.transposed_padding,
            lhs_dilation=self.stride,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
        )
        x = jnp.squeeze(x, axis=0)

        if self.bias is None:
            return x
        else:
            return x + self.bias()

    @property
    def transposed_padding(self) -> Sequence[Tuple[int, int]]:
        return [
            (d * (k - 1) - p[0], d * (k - 1) - p[1])
            for k, d, p in zip(
                self.kernel_size,
                self.dilation,
                self.padding,
            )
        ]



r"""Neural networks, layers and modules."""

from .activation import *  # noqa: F403
from .attention import *  # noqa: F403
from .container import *  # noqa: F403
from .dropout import *  # noqa: F403
from .einops import *  # noqa: F403
from .interpolation import *  # noqa: F403
from .linear import *  # noqa: F403
from .module import *  # noqa: F403
from .normalization import *  # noqa: F403
from .padding import *  # noqa: F403
from .pooling import *  # noqa: F403
from .recurrent import *  # noqa: F403
from .share import *  # noqa: F403
from .ssm import *  # noqa: F403
from .state import *  # noqa: F403



r"""Stateful modules

In Inox, in-place module mutations are not prohibited, but are not recommended as they
often lead to silent errors around JAX transformations. Instead, it is safer to
externalize the state of modules and handle mutations explicitely.

The :mod:`inox.nn.state` module provides a simple interface to declare the state of
modules and apply state updates. During initialization, mutable arrays are wrapped in
:class:`StateEntry` instances. After initialization, these arrays are pulled out and
replaced with hashable :class:`StateKey` instances using the :func:`export_state`
function. The state is represented by a dictionary which is used and updated during the
module's execution.

.. code-block:: python

    import inox
    import inox.nn as nn
    import jax
    import jax.numpy as jnp

    class Moments(nn.Module):
        def __init__(self, features):
            self.first = nn.StateEntry(jnp.zeros(features))
            self.second = nn.StateEntry(jnp.ones(features))

        def __call__(self, x, state):
            first = state[self.first]
            second = state[self.second]

            state = update_state(state, {
                self.first: 0.9 * first + 0.1 * x,
                self.second: 0.9 * second + 0.1 * x**2,
            })

            return state

    class MLP(nn.Module):
        def __init__(self, in_features, num_classes, key):
            keys = jax.random.split(key, 3)

            self.in_stats = Moments(in_features)
            self.out_stats = Moments(num_classes)

            self.l1 = nn.Linear(in_features, 64, key=keys[0])
            self.l2 = nn.Linear(64, 64, key=keys[1])
            self.l3 = nn.Linear(64, num_classes, key=keys[2])
            self.relu = nn.ReLU()

        def __call__(self, x, state):
            state = self.in_stats(x, state)

            x = self.l1(x)
            x = self.relu()
            x = self.l2(x)
            x = self.relu()
            x = self.l3(x)

            state = self.out_stats(x, state)

            return x, state

    key = jax.random.key(0)
    model = MLP(16, 3, key)
    model, state = nn.export_state(model)

    y, state = model(x, state)
"""

__all__ = [
    "StateEntry",
    "StateKey",
    "update_state",
    "export_state",
]

import jax.tree_util as jtu

from typing import Any, Dict, Hashable, NamedTuple, Tuple

import inox.tree


def is_entry(x: Any) -> bool:
    return isinstance(x, StateEntry)


class StateEntry(NamedTuple):
    r"""Wrapper to indicate a state entry.

    Arguments:
        value: A value.
    """

    value: Any


class StateKey(NamedTuple):
    r"""Wrapper to indicate a state key.

    Arguments:
        key: An hashable key.
    """

    key: Hashable

    def __repr__(self) -> str:
        return self.tree_repr()

    def tree_repr(self, **kwargs) -> str:
        return f"{self.__class__.__name__}({repr(self.key)})"


def update_state(state: Dict, mutation: Dict) -> Dict:
    r"""Creates a copy of the state dictionary and updates it.

    Arguments:
        state: The state dictionary.
        mutation: The update.

    Returns:
        The updated state dictionary.
    """

    state = state.copy()
    state.update(mutation)

    return state


def export_state(tree: inox.tree.PyTree) -> Tuple[inox.tree.PyTree, Dict]:
    r"""Pulls the state entries out of a tree.

    State entries are replaced by state keys which can be used to index the state
    dictionary.

    Arguments:
        tree: A tree or module.

    Returns:
        The stateless tree and the state dictionary.

    Example:
        >>> tree = {'a': 1, 'b': StateEntry(jax.numpy.zeros(2))}
        >>> tree, state = export_state(tree)
        >>> tree
        {'a': 1, 'b': StateKey("['b']")}
        >>> state
        {StateKey("['b']"): Array([0., 0.], dtype=float32)}
        >>> state[tree['b']]
        Array([0., 0.], dtype=float32)
    """

    nodes, treedef = jtu.tree_flatten_with_path(tree, is_entry)
    leaves = []
    state = {}

    for path, node in nodes:
        if is_entry(node):
            key = StateKey(jtu.keystr(path))
            node, state[key] = key, node.value

        leaves.append(node)

    tree = jtu.tree_unflatten(treedef, leaves)

    return tree, state



r"""Tests for the inox.nn module."""

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
import pickle
import pytest

from typing import Hashable

import inox.nn as nn

from inox import api
from inox.tree import is_array


def test_Module():
    module = nn.Module(
        a=0.0,
        b=nn.Parameter(jnp.ones(1)),
        c=[jnp.arange(2), "three"],
        d=False,
        e=nn.Module(f=np.zeros(5), g=range(6)),
        h=nn.ComplexParameter(7.0 + 8.0j),
    )

    # Flatten
    leaves, treedef = jtu.tree_flatten(module)

    assert not all(map(is_array, leaves))
    assert isinstance(treedef, Hashable)

    jtu.tree_unflatten(treedef, leaves)

    # Partition
    static, arrays = module.partition()

    assert all(map(is_array, arrays.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    static(arrays)

    ## filters (type)
    static, params, others = module.partition(nn.Parameter)

    assert all(key.endswith((".value", ".real", ".imag")) for key in params)
    assert all(map(is_array, params.values()))
    assert all(map(is_array, others.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    static(params, others)

    with pytest.raises(KeyError):
        static(params)

    with pytest.raises(KeyError):
        static(params, others, {"none": None})

    ## filters (callable)
    module.e.frozen = True

    static, frozen, others = module.partition(lambda x: getattr(x, "frozen", False))

    assert all(".e" in key for key in frozen)
    assert all(".e" not in key for key in others)
    assert all(map(is_array, frozen.values()))
    assert all(map(is_array, others.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    # Pickle
    data = pickle.dumps(module)
    pickle.loads(data)

    # Print
    assert repr(module)


@pytest.mark.parametrize("norm", ["layer", "group"])
def test_MLP(norm: str):
    key = jax.random.key(0)
    x = jax.random.normal(key, (1024, 3))
    y = jnp.linalg.norm(x, axis=-1, keepdims=True)

    class MLP(nn.Module):
        def __init__(self, key):
            keys = jax.random.split(key)

            self.l1 = nn.Linear(3, 64, key=keys[0])
            self.l2 = nn.Linear(64, 1, key=keys[1])
            self.relu = nn.ReLU()
            self.norm = nn.LayerNorm() if norm == "group" else nn.GroupNorm(4)

        def __call__(self, x):
            return self.l2(self.norm(self.relu(self.l1(x))))

    # __init__
    model = MLP(key)

    # __call__
    z0 = model(x[0])
    z1 = model(x)
    z2 = jax.vmap(model)(x)

    assert z0.shape == y[0].shape
    assert z1.shape == y.shape
    assert z2.shape == y.shape
    assert jnp.allclose(z1, z2)

    # JIT
    @api.jit
    def loss(model):
        return jnp.mean((model(x) - y) ** 2)

    loss(model)

    # Partition
    static, params, others = model.partition(nn.Parameter)

    assert all(key.endswith(".value") for key in params)
    assert all(map(is_array, params.values()))
    assert all(map(is_array, others.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    # Gradients
    grads = api.grad(lambda params: loss(static(params, others)))(params)
    params = jtu.tree_map(lambda x, y: x + y, params, grads)

    # Print
    assert repr(model)


def test_BatchNorm():
    key = jax.random.key(0)
    x = jax.random.uniform(key, (1024, 3))
    y = jnp.linalg.norm(x, axis=-1, keepdims=True)

    class MLP(nn.Module):
        def __init__(self, key):
            keys = jax.random.split(key)

            self.l1 = nn.Linear(3, 64, key=keys[0])
            self.l2 = nn.Linear(64, 1, key=keys[1])
            self.relu = nn.ReLU()
            self.norm = nn.BatchNorm(64)

        def __call__(self, x, state):
            x = self.l1(x)
            x = self.relu(x)
            x, state = self.norm(x, state)
            x = self.l2(x)

            return x, state

    # __init__
    model = MLP(key)
    model, state = nn.export_state(model)

    # __call__
    z, new = model(x, state)

    assert z.shape == y.shape
    assert not any(jtu.tree_leaves(jtu.tree_map(jnp.allclose, new, state)))

    with pytest.raises(TypeError):
        model(x, None)

    with pytest.raises(AssertionError):
        model(x[0], state)

    with pytest.raises(AssertionError):
        jax.vmap(model, in_axes=(0, None))(x, state)

    # JIT
    @api.jit
    def loss(model, state):
        z, state = model(x, state)

        return jnp.mean((z - y) ** 2), state

    _, state = loss(model, state)

    # Partition
    static, params, others = model.partition(nn.Parameter)

    assert all(key.endswith(".value") for key in params)
    assert all(map(is_array, params.values()))
    assert all(map(is_array, others.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    # Gradients
    def ell(params):
        return loss(static(params, others), state)

    grads, state = api.grad(ell, has_aux=True)(params)
    params = jtu.tree_map(lambda x, y: x + y, params, grads)

    # Print
    assert repr(model)


def test_share():
    key = jax.random.key(0)
    x = jax.random.uniform(key, (1024, 3))
    y = jnp.linalg.norm(x, axis=-1, keepdims=True)

    class MLP(nn.Scope):
        def __init__(self, key):
            keys = jax.random.split(key, 3)

            self.l1 = nn.Linear(in_features=3, out_features=64, key=keys[0])
            self.l2 = nn.Linear(in_features=64, out_features=64, key=keys[1])
            self.l4 = nn.Linear(in_features=64, out_features=1, key=keys[2])
            self.relu = nn.ReLU()

            self.l2 = nn.Reference("l2", self.l2)
            self.l2.cycle = self.l2
            self.l3 = self.l2
            self.l4.weight = nn.Reference("l4.weight", self.l4.weight)
            self.void = self.l4.weight

        def __call__(self, x):
            x = self.l1(x)
            x = self.l2(self.relu(x))
            x = self.l3(self.relu(x))
            x = self.l4(self.relu(x))

            return x

    # __init__
    model = MLP(key)

    # __call__
    z0 = model(x[0])
    z1 = model(x)
    z2 = jax.vmap(model)(x)

    assert z0.shape == y[0].shape
    assert z1.shape == y.shape
    assert z2.shape == y.shape
    assert jnp.allclose(z1, z2)

    # JIT
    @api.jit
    def loss(model):
        return jnp.mean((model(x) - y) ** 2)

    loss(model)

    # Partition
    static, params, others = model.partition(nn.Parameter)

    assert not any(key.startswith(".l3") or key.startswith(".void") for key in params)
    assert all(key.endswith(".value") for key in params)
    assert all(map(is_array, params.values()))
    assert all(map(is_array, others.values()))
    assert not any(map(is_array, jtu.tree_leaves(static)))

    # Gradients
    grads = api.grad(lambda params: loss(static(params, others)))(params)
    params = jtu.tree_map(lambda x, y: x + y, params, grads)

    # Print
    assert repr(model)



r"""Doctests configuration."""

import jax
import pytest

import inox


@pytest.fixture(autouse=True, scope="module")
def doctest_imports(doctest_namespace):
    doctest_namespace["jax"] = jax
    doctest_namespace["inox"] = inox
    doctest_namespace["nn"] = inox.nn
